#include "csurf.h"

void CSURF_advance_2(curve E, int n, const gfp p) {
  if (n < 0)
    fmpz_mod_neg(E->A, E->A, p);
  for (int i = n < 0 ? -n : n; i > 0; i--) {
    //                                     _______
    // A' : C' = 2( (2C-A)² - 2(6C + A - 4√C(2C+A))² ) : (2C-A)²

    fmpz_mod_add(tmp+0, E->C, E->C, p);
    fmpz_mod_add(tmp+2, tmp+0, E->C, p);
    fmpz_mod_mul(tmp+1, tmp+2, E->C, p);  // tmp1 = C(2C+A)
    fmpz_mod_sub(tmp+3, tmp+0, E->A, p);
    fmpz_mod_mul(E->C, tmp+3, tmp+3, p);  // C' = (2C-A)²
    fmpz_mod_add(tmp+3, tmp+2, tmp+0, p);
    fmpz_mod_add(tmp+2, tmp+3, tmp+0, p); // tmp2 = 6C+A
    
    fmpz_add_ui(tmp+3, p->n, 1);
    fmpz_tdiv_q_2exp(tmp+0, tmp+3, 2);
    fmpz_mod_pow_fmpz(tmp+3, tmp+1, tmp+0, p);
    fmpz_mod_add(tmp+0, tmp+3, tmp+3, p);    //          _______
    fmpz_mod_add(tmp+1, tmp+0, tmp+0, p);    // tmp1 = 4√C(2C+A)

    fmpz_mod_sub(tmp+0, tmp+2, tmp+1, p);
    fmpz_mod_mul(tmp+1, tmp+0, tmp+0, p);    //                     _______
    fmpz_mod_add(tmp+0, tmp+1, tmp+1, p);    // tmp0 = 2(6C + A - 4√C(2C+A))²

    fmpz_mod_sub(tmp+1, E->C, tmp+0, p);
    fmpz_mod_add(E->A, tmp+1, tmp+1, p);
  }
  if (n < 0)
    fmpz_mod_neg(E->A, E->A, p);
  curve_precomp(E, p);
}

int advance(curve E, point P, fmpz_t ord, fmpz_t cof, short *vec,
	    const gfp p, int threshold,
	    const unsigned short *fact, const unsigned char *bs_tab) {
  fmpz_t ord_P;
  point ker;
  int go = 0;

  fmpz_init_set(ord_P, ord);
  point_init(ker);
  
  // All degrees except 2, 3
  for (int i = CSIDH_LEN-1; i >= 2; i--) {
    int ell = fact[i];

    if (vec[i] > 0) {
      fmpz_divexact_ui(ord_P, ord_P, ell);
      point_mul(ker, P, E, ord_P, p);
      if (!fmpz_is_zero(ker->Z)) {
	if (ell < threshold)
	  VeluRenes(E, E, ker, ell, P, p);
	else
	  VeluResultant_set_bs(E, E, ker, ell, P, p, bs_tab[i]);
	vec[i] -= 1;
	if (vec[i] == 0) {
	  fmpz_divexact_ui(ord, ord, ell);
	  fmpz_mul_ui(cof, cof, ell);
	}
      }
    }
    if (vec[i] > 0)
      go = 1;
  }
  
  // Degree 3
  while (vec[1] > 0 && !fmpz_is_zero(P->Z)) {
    fmpz_divexact_ui(ord_P, ord_P, 3);
    point_mul(ker, P, E, ord_P, p);
    if (!fmpz_is_zero(ker->Z)) {
      VeluRenes(E, E, ker, 3, P, p);
      vec[1] -= 1;
      if (vec[1] == 0) {
	fmpz_divexact_ui(ord, ord, fact[1]);
	fmpz_mul_ui(cof, cof, fact[1]);
      }
    }
  }
  if (vec[1] > 0)
    go = 1;
  
  fmpz_clear(ord_P);
  point_clear(ker);

  return go;
}

void action(curve E1, const curve E, const short key[CSIDH_LEN],
	    const gfp p, flint_rand_t rng, int threshold,
	    const unsigned short *fact, const unsigned char *bs_tab) {
  short left[CSIDH_LEN], right[CSIDH_LEN];
  fmpz_t left_cof, left_ord, right_cof, right_ord;
  int go_left = 0, go_right = 0;
  point P;
  
  fmpz_init(left_cof); fmpz_init(left_ord);
  fmpz_init(right_cof); fmpz_init(right_ord);
  point_init(P);

  // Handle degree 2 isogenies
  curve_set(E1, E);
  if (fact == CSURF_fact) {
    CSURF_advance_2(E1, key[0], p);
  }
    
  // Handle other degrees
  fmpz_set_ui(left_cof, fact[0]);
  fmpz_set_ui(right_cof, fact[0]);
  left[0] = right[0] = 0;
  
  for (int i = 1; i < CSIDH_LEN; i++) {
    if (key[i] > 0) {
      left[i] = key[i];
      right[i] = 0;
      fmpz_mul_ui(right_cof, right_cof, fact[i]);
      go_left = 1;
    } else if (key[i] < 0) {
      left[i] = 0;
      right[i] = -key[i];
      fmpz_mul_ui(left_cof, left_cof, fact[i]);
      go_right = 1;
    } else {
      left[i] = right[i] = 0;
      fmpz_mul_ui(left_cof, left_cof, fact[i]);
      fmpz_mul_ui(right_cof, right_cof, fact[i]);
    }
  }
  fmpz_add_ui(tmp, p->n, 1);
  fmpz_divexact(left_ord,  tmp, left_cof);
  fmpz_divexact(right_ord, tmp, right_cof);

  while (go_left || go_right) {
    fmpz_randbits(P->X, rng, 60);
    fmpz_abs(P->X, P->X);
    fmpz_set_ui(P->Z, 1);
    int isoncurve = is_on_curve(P, E1, p);
    if (go_left && isoncurve) {
      point_mul(P, P, E1, left_cof, p);
      go_left = advance(E1, P, left_ord, left_cof, left, p, threshold, fact, bs_tab);
    } else if (go_right && !isoncurve) {
      point_mul(P, P, E1, right_cof, p);
      go_right = advance(E1, P, right_ord, right_cof, right, p, threshold, fact, bs_tab);
    }
  }
  
  point_clear(P);
  fmpz_clear(left_cof); fmpz_clear(left_ord);
  fmpz_clear(right_cof); fmpz_clear(right_ord);
}
