#include "csurf.h"

int is_on_curve(const point P, const curve E, const gfp p) {
  fmpz_mod_sub(tmp+0, P->X, P->Z, p);
  fmpz_mod_mul(tmp+1, tmp+0, tmp+0, p);  // (X - Z)²
  fmpz_mod_mul(tmp+0, P->X, P->Z, p);
  fmpz_mod_add(tmp+2, tmp+0, tmp+0, p);
  fmpz_mod_add(tmp+0, tmp+2, tmp+2, p);  // 4XZ
  fmpz_mod_mul(tmp+2, tmp+1, E->C, p);
  fmpz_mod_mul(tmp+3, tmp+0, E->A2C4, p);
  fmpz_mod_add(tmp+1, tmp+2, tmp+3, p);  // CX² + AXZ + CΖ²
  fmpz_mod_mul(tmp+2, tmp+0, E->C, p);
  fmpz_mod_mul(tmp+3, tmp+2, tmp+1, p);  // CXZ(CX² + AXZ + CΖ²)
  return is_square(tmp+3, p);
}

void point_double(point res, const point P, const curve E, const gfp p) {
  fmpz_mod_add(tmp+1, P->X, P->Z, p);
  fmpz_mod_mul(tmp+0, tmp+1, tmp+1, p);  // tmp0 = (X + Z)²
  fmpz_mod_sub(tmp+2, P->X, P->Z, p);
  fmpz_mod_mul(tmp+1, tmp+2, tmp+2, p);  // tmp1 = (X - Z)²
  fmpz_mod_mul(tmp+2, tmp+1, E->C, p);   // tmp2 = C (X - Z)²
  
  fmpz_mod_mul(res->X, tmp+2, tmp+0, p); // X = C (X² - Z²)²
  
  fmpz_mod_sub(tmp+3, tmp+0, tmp+1, p);  // tmp3 = 4 XZ
  fmpz_mod_mul(tmp+0, E->A2C4, tmp+3, p);// tmp0 = (A + 2C) XZ
  fmpz_mod_add(tmp+1, tmp+2, tmp+0, p);  // tmp1 = (CX² + AXZ + CZ²)
  
  fmpz_mod_mul(res->Z, tmp+1, tmp+3, p); // Z = 4 XZ (CX² + AXZ + CZ²)
}

void point_xadd(point res, const point P, const point Q, const point PmQ, const gfp p) {
  fmpz_mod_add(tmp+1, P->X, P->Z, p);
  fmpz_mod_sub(tmp+2, Q->X, Q->Z, p);
  fmpz_mod_mul(tmp+0, tmp+2, tmp+1, p);  // tmp0 = (Q.X - Q.Z) (P.X + P.Z)

  fmpz_mod_sub(tmp+2, P->X, P->Z, p);
  fmpz_mod_add(tmp+3, Q->X, Q->Z, p);
  fmpz_mod_mul(tmp+1, tmp+2, tmp+3, p);  // tmp1 = (P.X - P.Z) (Q.X + Q.Z)

  fmpz_mod_add(tmp+2, tmp+0, tmp+1, p);
  fmpz_mod_mul(tmp+3, tmp+2, tmp+2, p);  // tmp3 = (tmp0 + tmp1)² = 4 (Q.X P.X - Q.Z P.Z)²
  fmpz_mod_sub(tmp+2, tmp+0, tmp+1, p);
  fmpz_mod_mul(tmp+0, tmp+2, tmp+2, p);  // tmp0 = (tmp0 - tmp1)² = 4 (Q.X P.Z - Q.Z P.X)²

  fmpz_mod_mul(tmp+1, tmp+3, PmQ->Z, p);  // res.X = tmp3 PmQ.Z
  fmpz_mod_mul(tmp+2, tmp+0, PmQ->X, p);  // res.Z = tmp0 PmQ.X

  fmpz_set(res->X, tmp+1);
  fmpz_set(res->Z, tmp+2);
}

void point_mul(point res, const point P, const curve E, const fmpz_t n, const gfp p) {
  point x0, x1;
  point_init(x0); point_init(x1);

  point_set(x0, P);
  point_double(x1, P, E, p);
  for (long i = fmpz_bits(n) - 2; i >= 0; i--) {
    if (fmpz_tstbit(n, i)) {
      point_xadd(x0, x0, x1, P, p);
      point_double(x1, x1, E, p);
    } else {
      point_xadd(x1, x0, x1, P, p);
      point_double(x0, x0, E, p);
    }
  }
  point_set(res, x0);
  
  point_clear(x0); point_clear(x1);
}
