#include "csurf.h"

fmpz tmp[TMPS_NUM];

int main() {
  init_tmps();
  
  flint_rand_t rng;
  flint_randinit(rng);
  
  fmpz_mod_ctx_t p; curve E, E1, E2; point P, PP, Q, ker;
  CSURF_ctx(p);
  curve_init(E); curve_init(E1); curve_init(E2);
  CSURF_start(E, p);
  point_init(P);
  point_init(PP);
  point_init(Q);
  point_init(ker);

  fmpz_t mul, cof;
  fmpz_init(mul); fmpz_init(cof);
  for (int i = 1; i < CSIDH_LEN; i++) {
    fmpz_randm(P->X, rng, p->n);
    fmpz_set_ui(P->Z, 1);

    fmpz_add_ui(cof, p->n, 1);
    fmpz_divexact_ui(cof, cof, CSURF_fact[i]);

    point_mul(ker, P, E, cof, p);
    if (fmpz_is_zero(ker->Z))
      continue;

    fmpz_set_ui(mul, CSURF_fact[i]);
    point_mul(Q, ker, E, mul, p);
    assert(fmpz_is_zero(Q->Z));

    fmpz_set_ui(mul, i+2);
    point_mul(Q, P, E, mul, p);
    point_set(PP, P);
    VeluRenes(E1, E, ker, CSURF_fact[i], P, p);
    if (i > 1) {
      VeluResultant(E2, E, ker, CSURF_fact[i], PP, p);
      assert(curve_equal(E2, E1, p));
      assert(point_equal(PP, P, p));
    }
    
    VeluRenes(E, E, ker, CSURF_fact[i], Q, p);
    assert(curve_equal(E, E1, p));
    point_mul(P, P, E, mul, p);
    assert(point_equal(P, Q, p));

    point_mul(P, P, E, cof, p);
    assert(fmpz_is_zero(P->Z));
  }

  printf("    \033[1;32mAll tests passed\033[0m\n");
  exit(0);
}
