#include "csurf.h"

fmpz tmp[TMPS_NUM];

int main() {
  init_tmps();
  
  flint_rand_t rng;
  flint_randinit(rng);
  
  fmpz_mod_ctx_t p; curve E; point P, Q;
  CSURF_ctx(p);
  curve_init(E);
  CSURF_start(E, p);
  point_init(P);
  point_init(Q);

  for (int i = 0; i < 10; i++) {
    fmpz_randm(P->X, rng, p->n);
    fmpz_set_ui(P->Z, 1);
    for (int j = 0; j < 10; j++) {
      point_mul(Q, P, E, p->n, p);
      assert(point_equal(Q, P, p));
    }
  }

  fmpz_set_ui(E->A, 6);
  curve_precomp(E, p);
  
  for (int i = 0; i < 10; i++) {
    fmpz_randm(P->X, rng, p->n);
    fmpz_set_ui(P->Z, 1);
    for (int j = 0; j < 10; j++) {
      point_mul(Q, P, E, p->n, p);
      assert(point_equal(Q, P, p));
    }
  }

  // Elligator2 formulas
  for (int i = 0; i < 10; i++) {
    fmpz_randm(tmp+1, rng, p->n);         // u
    if (fmpz_is_zero(tmp+1))
      continue;
    fmpz_mod_mul(tmp+0, tmp+1, tmp+1, p); // u²
    fmpz_sub_ui(tmp+1, tmp+0, 1);         // u² - 1
    if (fmpz_is_zero(tmp+1))
      continue;
    fmpz_mod_mul(tmp+1, tmp+1, E->C, p);  // C (u² - 1)
    fmpz_set(P->Z, tmp+1); fmpz_set(Q->Z, tmp+1);
    fmpz_set(P->X, E->A);
    fmpz_mod_mul(Q->X, E->A, tmp+0, p);
    fmpz_mod_neg(Q->X, Q->X, p);          // -A u²
    assert(is_on_curve(P, E, p) != is_on_curve(Q, E, p));
  }
  
  printf("    \033[1;32mAll tests passed\033[0m\n");
  exit(0);
}
