#!/usr/bin/env python

import argparse, io, pandas as pd, subprocess as sp

pd.set_option('display.max_rows', 500)
pd.set_option('display.max_columns', 40)

def parse(csv_string, names, sep="\s+", comment="#"):
    return pd.read_csv(io.StringIO(csv_string), 
                       sep=sep, comment=comment, names=names)


degrees = [5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79,
   83, 89, 97, 101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 151, 157, 163, 167,
   173, 179, 181, 191, 193, 197, 199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 263,
   269, 271, 277, 281, 283, 293, 307, 311, 313, 317, 331, 337, 349, 353, 367, 373, 379,
   383, 389]

parser = argparse.ArgumentParser('Tune # of babysteps')
parser.add_argument('-e', type=int, default=20, help='# of evaluations')
parser.add_argument('-s', type=int, default=20, help='# of samples')
parser.add_argument('-a', choices=['CSIDH', 'CSURF'], default='CSIDH', help='algorithm')
args = parser.parse_args()
    
tunes = [0, 0]
for d in degrees:
    r = sp.run(["./bench-velu.exe", "-e%d" % args.e, "-s%d" % args.s, "-l%d" % d, "-p%s" % args.a],
                   stdout=sp.PIPE)
    med = parse(str(r.stdout.decode()), ['p', 'c', 't']).groupby('p').median()
    tunes.append(med.c.idxmin())

print(tunes)
